
#ifndef SHARE_VM_GC_SHENANDOAH_SHENANDOAHROOTPROCESSOR_HPP
#define SHARE_VM_GC_SHENANDOAH_SHENANDOAHROOTPROCESSOR_HPP

#include "gc/shared/strongRootsScope.hpp"
#include "memory/allocation.hpp"
#include "runtime/mutex.hpp"

class CLDClosure;
class CodeBlobClosure;
class G1CollectedHeap;
class G1GCPhaseTimes;
class G1ParPushHeapRSClosure;
class Monitor;
class OopClosure;
class SubTasksDone;

class ShenandoahRootProcessor : public StackObj {
  SubTasksDone* _process_strong_tasks;
  StrongRootsScope _srs;

  enum Shenandoah_process_roots_tasks {
    SHENANDOAH_RP_PS_Universe_oops_do,
    SHENANDOAH_RP_PS_JNIHandles_oops_do,
    SHENANDOAH_RP_PS_JNIHandles_weak_oops_do,
    SHENANDOAH_RP_PS_ObjectSynchronizer_oops_do,
    SHENANDOAH_RP_PS_FlatProfiler_oops_do,
    SHENANDOAH_RP_PS_Management_oops_do,
    SHENANDOAH_RP_PS_SystemDictionary_oops_do,
    SHENANDOAH_RP_PS_ClassLoaderDataGraph_oops_do,
    SHENANDOAH_RP_PS_jvmti_oops_do,
    SHENANDOAH_RP_PS_CodeCache_oops_do,
    SHENANDOAH_RP_PS_filter_satb_buffers,
    SHENANDOAH_RP_PS_refProcessor_oops_do,
    // Leave this one last.
    SHENANDOAH_RP_PS_NumElements
  };

  void process_java_roots(OopClosure* scan_non_heap_roots,
                          CLDClosure* thread_stack_clds,
                          CLDClosure* scan_strong_clds,
                          CLDClosure* scan_weak_clds,
                          CodeBlobClosure* scan_strong_code,
                          uint worker_i);

  void process_vm_roots(OopClosure* scan_non_heap_roots,
                        OopClosure* scan_non_heap_weak_roots,
                        uint worker_i);

public:
  ShenandoahRootProcessor(ShenandoahHeap* heap, uint n_workers);

  void process_roots(OopClosure* strong_oops,
		     OopClosure* weak_oops,
		     CLDClosure* strong_clds,
		     CLDClosure* weak_clds,
		     CLDClosure* thread_stack_clds,
		     CodeBlobClosure* strong_code);

  // Apply oops, clds and blobs to all strongly reachable roots in the system
  void process_strong_roots(OopClosure* oops,
                            CLDClosure* clds,
                            CodeBlobClosure* blobs);

  // Apply oops, clds and blobs to strongly and weakly reachable roots in the system
  void process_all_roots(OopClosure* oops,
                         CLDClosure* clds,
                         CodeBlobClosure* blobs);

  // Number of worker threads used by the root processor.
  uint n_workers() const;
};

#endif // SHARE_VM_GC_SHENANDOAH_SHENANDOAHROOTPROCESSOR_HPP
